from __future__ import annotations

import math
from random import Random

import numpy as np


def uniform_qualitle_sampling(
    population: list[tuple[float, str]],
    n_sample: int,
    rng: Random,
    shuffle: bool = True,
) -> list[str]:
    """Sample SMILES by sampling uniformly from logarithmically spaced top-N."""

    samples: list[str] = []
    quantiles = 1 - np.logspace(-3, 0, 25)
    n_samples_per_quanitile = int(math.ceil(n_sample / len(quantiles)))
    for q in quantiles:
        score_threshold = np.quantile([s for s, _ in population], q)
        eligible_population = [smiles for score, smiles in population if score >= score_threshold]
        samples.extend(rng.choices(population=eligible_population, k=n_samples_per_quanitile))

    # Shuffle samples to decrease correlations between adjacent samples
    if shuffle:
        rng.shuffle(samples)

    return samples[:n_sample]  # in case there are slightly too many samples



""" Main code for running GAs. """
from __future__ import annotations

import logging
import random
from dataclasses import dataclass
from pprint import pformat
from typing import Any, Callable, Optional, Union

import joblib
import numpy as np

from .cached_function import CachedBatchFunction

# Logger with standard handler
ga_logger = logging.getLogger(__name__)


@dataclass
class GAResults:
    """Results from a GA run."""

    population: list[tuple[float, str]]
    scoring_func_evals: dict[str, float]
    gen_info: list[dict[str, Any]]


def run_ga_maximization(
    *,
    scoring_func: Union[Callable[[list[str]], list[float]], CachedBatchFunction],
    starting_population_smiles: set[str],
    sampling_func: Callable[[list[tuple[float, str]], int, random.Random], list[str]],
    offspring_gen_func: Callable[[list[str], int, random.Random, Optional[joblib.Parallel]], set[str]],
    selection_func: Callable[[int, list[tuple[float, str]]], list[tuple[float, str]]],
    max_generations: int,
    population_size: int,
    offspring_size: int,
    rng: Optional[random.Random] = None,
    num_samples_per_generation: Optional[int] = None,
    logger: Optional[logging.Logger] = None,
    parallel: Optional[joblib.Parallel] = None,
) -> GAResults:
    """
    Runs a genetic algorithm to maximize `scoring_func`.

    Args:
        scoring_func: Function that takes a list of SMILES and returns a list of scores.
            Convention: scoring functionis being MAXIMIZED!
        starting_population_smiles: Set of SMILES to start the GA with.
        sampling_func: Function that takes a list of (score, smiles) tuples and returns a list of SMILES.
            This is used to sample SMILES from the population to create offspring.
        offspring_gen_func: Function that takes a list of SMILES and returns a list of SMILES.
            This can be anything (e.g. crossover, mutation).
        selection_func: Function to select the new population.
            This can be anything from "take the N best scores" to some more complicated method to pick
            a diverse subset.
        max_generations: Maximum number of generations to run the GA for.
        population_size: Number of SMILES to keep in the population.
        offspring_size: Number of offspring to create per generation.
        rng: Random number generator.
        num_samples_per_generation: Number of samples to take from the population to create offspring.
        logger: Logger to use.
        parallel: Joblib parallel object to use (to generate offspring in parallel).

    Returns:
        GAResults object containing the population, scoring function, and information about each generation.
    """

    # ============================================================
    # 0: Process input variables
    # ============================================================
    logger = logger or ga_logger
    logger.info("Starting GA maximization...")
    num_samples_per_generation = num_samples_per_generation or 2 * offspring_size
    rng = rng or random.Random()

    # Create the cached scoring function
    if not isinstance(scoring_func, CachedBatchFunction):
        scoring_func = CachedBatchFunction(scoring_func)
    start_cache_size = len(scoring_func.cache)
    logger.debug(f"Starting cache made, has size {start_cache_size}")

    # ============================================================
    # 1: prepare initial population
    # ============================================================

    # Score initial SMILES
    population_smiles = list(starting_population_smiles)
    population_scores = scoring_func.eval_batch(population_smiles)
    _starting_max_score = max(population_scores)
    logger.debug(f"Initial population scoring done. Pop size={len(population_smiles)}, Max={_starting_max_score}")
    population = list(zip(population_scores, population_smiles))
    del population_scores, population_smiles, _starting_max_score

    # Perform initial selection
    population = selection_func(population_size, population)

    # ============================================================
    # 2: run GA iterations
    # ============================================================

    # Run GA
    gen_info: list[dict[str, Any]] = []
    for generation in range(max_generations):
        logger.info(f"Start generation {generation}")

        # Separate out into SMILES and scores
        _, population_smiles = tuple(zip(*population))  # type: ignore[assignment]

        # Sample SMILES from population to create offspring
        samples_from_population = sampling_func(population, num_samples_per_generation, rng)

        # Create the offspring
        offspring = offspring_gen_func(
            samples_from_population,
            offspring_size,
            rng,
            parallel,
        )

        # Add to population, ensuring uniqueness
        population_smiles = list(set(population_smiles) | offspring)  # type: ignore[misc]  # thinks var deleted
        logger.debug(f"\t{len(offspring)} created")
        logger.debug(f"\tNew population size = {len(population_smiles)}")
        del offspring

        # Score new population
        logger.debug("\tCalling scoring function...")
        population_scores = scoring_func.eval_batch(population_smiles)
        logger.debug(f"\tScoring done, best score now {max(population_scores)}.")

        # Select new population
        population = list(zip(population_scores, population_smiles))
        population = selection_func(population_size, population)

        # Log results of this generation
        population_scores, population_smiles = tuple(zip(*population))  # type: ignore[assignment]
        gen_stats_dict = dict(
            max=np.max(population_scores),
            avg=np.mean(population_scores),
            median=np.median(population_scores),
            min=np.min(population_scores),
            std=np.std(population_scores),
            size=len(population_scores),
            num_func_eval=len(scoring_func.cache) - start_cache_size,
        )
        logger.info("End of generation. Stats:\n" + pformat(gen_stats_dict))
        gen_info.append(gen_stats_dict)
        del population_scores, population_smiles

    # ============================================================
    # 3: Create return object
    # ============================================================
    logger.info("End of GA. Returning results.")
    return GAResults(population=population, scoring_func_evals=scoring_func.cache, gen_info=gen_info)



""" Code for functions which cache inputs for efficiency. """
from __future__ import annotations

from typing import Callable, Optional


class CachedFunction:
    """Function which caches previously computed values to avoid repeat computation."""

    def __init__(
        self,
        f: Callable,
        cache: Optional[dict] = None,
        transform: Optional[Callable] = None,
        max_cache_size: Optional[int] = None,
    ):
        """Init function

        :param f: The function to cache
        :type f: callable
        :param cache: dict mapping known inputs-> outputs of f, defaults to None
        :type cache: dict, optional
        :param transform: optional transform function to apply to values
            of f (e.g. scaling output to be in [0, 1]), defaults to None
        :type transform: callable, optional
        """
        self._f = f
        self.cache = cache or dict()
        self.transform = transform
        self.max_cache_size = max_cache_size
        self._trim_cache()

    def _batch_f_eval(self, input_list):
        return [self._f(x) for x in input_list]

    def _batch_transform(self, output_list):
        if self.transform is None:
            return output_list
        else:
            return [self.transform(x) for x in output_list]

    def _trim_cache(self) -> None:
        """Clears the cache if it is too large."""
        if self.max_cache_size is not None and len(self.cache) > self.max_cache_size:
            self.cache.clear()

    def eval_batch(self, inputs):
        # Eval function at non-cached inputs
        inputs_not_cached = [x for x in inputs if x not in self.cache]
        outputs_not_cached = self._batch_f_eval(inputs_not_cached)

        # Add new values to cache
        for x, y in zip(inputs_not_cached, outputs_not_cached):
            self.cache[x] = y

        # Get and transform outputs
        outputs = [self.cache[x] for x in inputs]
        outputs = self._batch_transform(outputs)
        return outputs

    def eval_non_batch(self, inputs):
        inputs = [inputs]
        outputs = self.eval_batch(inputs)
        assert len(outputs) == 1
        outputs = outputs[0]
        return outputs

    def __call__(self, inputs, batch=True):
        # Ensure it is in batch form
        if batch:
            output = self.eval_batch(inputs)
        else:
            output = self.eval_non_batch(inputs)

        self._trim_cache()
        return output


class CachedBatchFunction(CachedFunction):
    """
    Special kind of cached function where f takes a batch of inputs
    instead of a single input. The input to the underlying function
    (and transform function) will always be a batch of data.
    """

    def _batch_f_eval(self, input_list):
        return self._f(input_list)



from __future__ import annotations

import joblib
import pytest
from rdkit import Chem
from rdkit.Chem import QED

from mol_ga import default_ga
from mol_ga.mol_libraries import random_zinc


def qed_value(smiles: str) -> float:
    """Calculate QED value of a SMILES string."""
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return -1.0
    return QED.qed(mol)


def batch_qed(smiles: list[str]) -> list[float]:
    """Calculate QED values of a list of SMILES strings."""
    return [qed_value(s) for s in smiles]


@pytest.mark.parametrize("parallel", [False, True])
def test_smoke(parallel: bool):
    """Test that the default GA runs with and without parallel."""
    kwargs = dict(
        starting_population_smiles=random_zinc(1000),
        scoring_function=batch_qed,
        max_generations=10,
        offspring_size=100,
    )

    if parallel:
        with joblib.Parallel(n_jobs=-1) as parallel:
            output = default_ga(**kwargs, parallel=parallel)
    else:
        output = default_ga(**kwargs, parallel=None)

    assert max(output.population)[0] > 0.935
